#include "palsol.h"
#include "palerr.h"

/*------------------------------------------------
 * PALSol class constructor.                 
 *
 * INPUT:
 * 	n	size of QEP
 * 	p	order of Pade approximant
 * 	rk	rank of the damping matrix
 *	sigma	shift applied by PAL method
 *	nev	number of eigenvalue required
 *	M,D,K	mass, damping, stiffness matrix
 *----------------------------------------------*/

PALSol::
PALSol(	int pp, COMPLEX sigmap, int nevp, 
	PALMatrix Mp, PALMatrix Dp, PALMatrix Kp):
	M(Mp),	D(Dp),	K(Kp)
{

	/* A) Data assignment */

	p 	=	pp;
	sigma	=	sigmap;
	nev	=	nevp;
	D.LowRankDcomp(E, F, rk);
	n 	= 	M.m; 
	nl	= 	n + p*rk;

	/* B) Compute Pade parameter a and b. */

 	const double PI = 3.141592653589793238462;
  	b 	= 	new double[p];
  	a 	= 	new double[p];

  	for(int i=0; i<p; i++){
    		b[ i ] = cos( (i + 1) * PI / (2 * p + 1) );
    		a[ i ] = sin( (i + 1) * PI / (2 * p + 1) );
    		a[ i ] = a[ i ] / b[ i ] * sqrt(2) / sqrt(2 * p + 1);
    		b[ i ] = b[ i ] * b[ i ];
  	}

	/* C) Parameter Initialization */

	threshold = 	0.1;	
  	rtol	= 	1.0E-16;
  	artol	= 	1.0E-16;
  	maxit	= 	300;
	nconv	=	-1;
	delta   =	1.0;
	nzmax	=	M.nnz+D.nnz+K.nnz;
  	order	= 	1;	
	ncv     = 	2*nev + 1;

	isscaling = iseigv = isresid = true;
	isfactor= eigdone = resdone = isscaled = false;

	normm = normd = normk = 0;

	isv0 	=	0; // random starting vector
}



/* parameter checking */
bool PALSol::
parachk(){

	/* matrix size checking */

	if( M.m!= M.n || D.m!=D.n || K.m!=K.n||
		M.m!=D.m || M.m!=K.m ||  n<0){
		erprint("Matrices M, D, K and K are with invalid/inconsistant size!");
		return false;
	}

	if( p < 1 ){
		erprint("Invalid Pade order p!");
		return false;
	}

	if( ABS(sigma) == 0 ){
		erprint("Shift sigma must not be zero!");
		return false;
	}

	return true;
}



double PALSol::
scaling() // compute norm estimate and scaling factors
{
	if(isscaled || !isscaling) return delta; // precheck

    // compute matrix norms
	normm = M.norm1();
  	normd = D.norm1();
  	normk = K.norm1();
      
    double t; // internal variable


  	// compute scaling factors delta= 1/ max(sigma^2*nm, 2*p*sigma*nd, nk)
  	delta 	= ABS(sigma)*ABS(sigma) * normm;
 	t	= 2*p*normd*ABS(sigma);
  	delta 	= (delta<t)? t : delta;
  	delta 	= (delta<normk)? normk : delta;
  	delta 	= 1.0/delta;

  	isscaled = true;

  	return delta;
}



void PALSol::
Qsigma(COMPLEX* val, int* colptr, int* rowind, int &nnz, int &info)
{ 
  	int 	pm, pd, pk, mdx, idxmh, idxdh, idxkh, // internal variables
	      	idxmt, idxkt, idxdt; 

	info 	= 0;
	nnz 	= 0;

	/* Compute Q(sigma) column by column */

	for(int i=0; i<=n-1; i++){

		idxmh = M.colptr[i];
	   	idxmt = M.colptr[i+1];	// pointers to column i and i+1, matrix M
	   	idxdh = D.colptr[i];
	   	idxdt = D.colptr[i+1];	// pointers to column i and i+1, matrix D
	   	idxkh = K.colptr[i];
	   	idxkt = K.colptr[i+1];	// pointers to column i and i+1, matrix K

	   	colptr[i] = nnz; 		// column pointers to ith column of Q

	   	while( true ){

			pm = ( idxmh == idxmt )? n : M.rowind[ idxmh ];
					// unvisited row index of M in column i
			pd = ( idxdh == idxdt )? n : D.rowind[ idxdh ];
					// unvisited row index of D in column i
			pk = ( idxkh == idxkt )? n : K.rowind[ idxkh ];
					// unvisited row index of K in column i

	     		mdx = ( pm < pd )? pm : pd;
	     		mdx = ( mdx < pk )? mdx : pk;
					// find the minimum index among pm, pd, pk
	
			if( mdx == n ) break;
					// i-th column completed

			rowind[ nnz ] = mdx;	// updates row index of Q
     	
			val[ nnz ] 	= 0;
			if( mdx == pm ){
			       	val[ nnz ] = val[ nnz ] + sigma * sigma * M.val[ idxmh ];
		       		idxmh++;
			}
	       		if( mdx == pd){
			       	val[ nnz ] = val[ nnz ] + sigma * D.val[ idxdh ];
				idxdh++;
			}
		      	if( mdx == pk){
			       	val[ nnz ] = val[ nnz ] + K.val[ idxkh ]; 
				idxkh++;
			} 

		nnz = nnz + 1;			// updates nnzq
	   	}
	}

	colptr[ n ] = nnz; 		// updating last colptr
	
	if( nnz > nzmax ) info = -1;

	return;
}// compute Q(sigma)  



void PALSol::
lufactor() // compute LU factorization of Q(s)
{

	/* reserve storage for Q(s), in CSR format 	*/
       
    COMPLEX	*matms = new COMPLEX[ nzmax ];
	int    	*rowindms = new int[ nzmax ];
	int   	*colptrms = new int[ n+1 ];
	int	nnzms, info;


	/* compute Q(sigma)				*/

 	Qsigma(matms, colptrms, rowindms, nnzms, info);	

	if( info == -1 ){
		erprint("nnz of Q(sigma) exceeds nzmax!");
	}

	/* prepare data for SUPERLU.			*/

  	SuperMatrix 	A;	
  	zCreate_CompCol_Matrix(&A, n, n, nnzms, (doublecomplex *)matms,
                         rowindms, colptrms, SLU_NC, SLU_Z, SLU_GE);

	/*
	 * LU factorization to A 
	 */

	/* Defining local variables. */

  	int         	*etree;
  	SuperMatrix 	AC;

  	/* Setting default values for gstrf parameters. */

  	double	drop_tol  	= 0.0;
  	int    	panel_size 	= sp_ienv( 1 );
  	int    	relax      	= sp_ienv( 2 );
  	superlu_options_t 	options;

      	set_default_options( &options );
  	options.DiagPivotThresh = threshold;

	 /* Reserving memory for etree (used in matrix decomposition). */

  	etree = new int[ n ];
  	permc = new int[ n ];
  	permr = new int[ n ];

  	StatInit( &stat);		//StatInit(panel_size, relax);

 	get_perm_c(order, &A, permc);	// Defining the column permutation of matrix A
  					// (using minimum degree ordering on A'*A).
  	sp_preorder(&options, &A, permc, etree, &AC);

  	zgstrf(&options, &AC, drop_tol, relax, panel_size, etree,
        	NULL, 0, permc, permr, &L, &U, &stat, &info);

    isfactor = true;

  	/* Deleting AC and etree, and working space */
	
    Destroy_CompCol_Permuted( &AC );
  	delete[] 	matms;
  	delete[] 	rowindms;
  	delete[] 	colptrms;
  	delete[] 	etree;

  // POSSIBLE IMPROVEMENTS: check info...
}



void PALSol::  // compute w = op(v); 
MultMv(COMPLEX* v, COMPLEX* w)
{

	COMPLEX zzero(0.0,0.0), alpha;
  	int             idx;
        COMPLEX* temp  = new COMPLEX[rk]; // internal working space

  	for(int i=0;i<n+rk*p;i++) w[i] =  zzero;	// initialize w to all zero;

  	/* A) compute w(1:n) =  Q^-1 (-sigma^2 M*v(1:n)-sqrt(sigma)F(Ir\cron a^T*E)v(n+1:end) ) */

    /* A.1) compute w(1:n) = -sigma^2*M*v(1:n) */

  	alpha = -sigma*sigma;
  	M.MultMv(v,w, alpha);

    /* A.2) compute temp = (I_r\cron a^T*E)*v(n+1:end) */

  	for(int i=0;i<rk; i++){
    		idx = n+i*p;
    		temp[i]=-a[0]*b[0]*v[idx];
    		for(int j=1;j<p; j++){
       			temp[i] = temp[i] -a[j]*b[j]*v[idx+j]; 
    		}
  	}

    /* A.3) compute w(1:n) = w(1:n) -sqrt(sigma) F * temp. */

  	alpha = -sqrt(sigma/delta);
  	E.MultMv(temp,w,alpha);  

  	/* B) compute Q^{-1} w(1:n) */

	if(!isfactor) lufactor(); 	// call LU factorization if not factored.

  	int         info, one=1;
  	SuperMatrix B;
  	zCreate_Dense_Matrix(&B, n, one, (doublecomplex *)w, n, SLU_DN, SLU_Z, SLU_GE);
  	StatInit(&stat);
  	trans_t trans = NOTRANS;
  	zgstrs(trans, &L, &U, permc, permr, &B, &stat, &info);
  	Destroy_SuperMatrix_Store(&B);

  	/* C) compute w(n+1:end) = (Ir\cron E)*v(n+1:end) - sqrt(sigma)(Ir\cron a)F^T w(1:n) */

  	for(int i=0;i<rk;i++) temp[i]=zzero;
  	alpha = 1.0;

    /* C.1) compute temp=F^T*w(1:n) */
  
 	F.MultMv(w,temp,alpha);	

    /* C.2) compute w2 = sqrt(sigma)(I_r\cron a)*temp. */

	for(int i=0;i<rk;i++){ 
		for(int j = 0;j<p;j++){
      			w[n+i*p+j] = -b[j]*v[n+i*p+j]-sqrt(sigma*delta)*a[j]*temp[i];
   	 	}
  	}

  	return;
}




void PALSol::  // Eigenvalue computation using ARPACK
EigComp()
{
	/* PARAMETER AND WORKSPACE SETTING UP */

	int	ido  = 0;		// First call to ARPACK
	char	bmat = 'I'; // Bmat
        std::string which("LM");	// Part of spectral

	int	lworkl = ncv * ( 3 * ncv + 6 );	
					        // Dimension of internal vectors
	int	lworkv = 2 * ncv;	// Dimension of internal vectors
	int	lrwork = ncv;		// Dimension of internal vectors
	COMPLEX *V     = new COMPLEX[ nl * ncv + 1 ];	
					// Arnoldi vectors
	COMPLEX *resid = new COMPLEX[ nl ];
					// Working space
	COMPLEX *workd = new COMPLEX[ 3 * nl + 1 ];
					// Working space 
	COMPLEX *workl = new COMPLEX[ lworkl + 1 ];
					// Working space
	COMPLEX *workv = new COMPLEX[ lworkv + 1 ];
					// Working space
	double 	*rwork = new double[ lrwork + 1];
					// Working space
	
	int		iparam[ 12 ];	// Vector that handles ARPACK parameters.
	int		ipntr[ 15 ];	// Vector than handles ARPACK pointers.

	nconv		= 0;		// No eigenvalues found yet
	iparam[ 1 ]	= 1;		// Using AutoShift: CAN USER-DEFINED
	iparam[ 3 ] = maxit;	// Maximum number of IRAM
	iparam[ 4 ] = 1;		// Block size must be 1.
	iparam[ 7 ] = 1;		// mode bmat=1.

	int	info 	= isv0; // Use random starting vectors if isv0 = 0,
       					// otherwise user-defined.

	scaling();			// Call scaling function,
       					// it will do nothing if isscaling = false.

    if ( !isfactor ) lufactor();	// LU factorization, if LU is not done.

	bool	BasisOK = false;

	/* MAIN ARPACK LOOP: REVERSE COMMUNICATION */

	while ( !BasisOK ) {
	
		/* Call ARPACK routine for updating Arnoldi basis */

		znaupd_( &ido, &bmat, &nl, which.c_str(), &nev, &artol, resid, &ncv,
			&V[1], &nl, &iparam[1], &ipntr[1], &workd[1],
			&workl[1], &lworkl, &rwork[1], &info);

		if ( ido == 99 ) {
			nconv = iparam[ 5 ];
			if ( info >= 0 ) BasisOK = true;
		}

		if (( ido == -1 ) || ( ido == 1 )) {
			// Performing Matrix vector multiplication: y<- op*x.
			MultMv( &workd[ ipntr[1] ], &workd[ ipntr[2] ]);
		}

	}

	/* Check _naupd error message. */

	if( info <0 ){ 
		auperr(info);
		erprint("ARPACK routine not successful"); 
	}

	/* Computing Ritz values */

	EigVal = new COMPLEX[ nev + 1 ]; 	// Eigenvalues
	EigVec = &V[ 1 ]; 			        // Eigenvectors

	int	rvec = (iseigv) ? 1 : 0;	// 0 not to compute eigenvectors,
						            // 1 compute eigenvectors.
						
	char	HowMny = 'A';			// 'A' for Ritz vectors, 
						            // 'P' for Shur vectors.
	
	int	*iselect = new int[ ncv ];	// Internal working space.

	/* Call ARPACK routine for the Ritz values and/or Ritz vectors. */

	zneupd_( &rvec, &HowMny, iselect, EigVal, EigVec, &nl, &sigma, &workv[1], 
		 &bmat, &nl, which.c_str(), &nev, &artol, resid, &ncv, 
		 &V[1], &nl, &iparam[1], &ipntr[1], &workd[1], &workl[1], 
		 &lworkl, &rwork[1], &info);

	/* Check _neupd error message. 			*/
	if( info <0 ){ 
		euperr(info);
		erprint("ARPACK routine not successful"); 
	}

	/* Convert LEP eigenvalues back to QEP.		*/
	for( int i = 0; i < nev; i++) {
		EigVal[ i ] = sigma * sqrt( 1.0 / EigVal[ i ] + 1.0 );
	}

	eigdone = true;		// Eigenvalue computation complete.
	
	/* Release storage.				*/
	delete[] resid;
	delete[] workd;
	delete[] workl;
	delete[] workv;
	delete[] rwork;

	/* Return if only eigenvalues are required. 	*/
	if(!iseigv) return; 		

	/* Normalize the eigenvectors. 			*/
	double 	tnm;
	int	inc=1;
	for ( int i=0; i<nev; i++ ) {
		tnm = dznrm2(&n, &EigVec[ i*nl ], &inc);
		tnm = 1.0 / tnm;
		zdscal(&n, &tnm, &EigVec[ i*nl ], &inc);
	}

	/* Return if relative residual norms are not required */
	if(!isresid) return;
	
	ResidComp();		// Relative residual norms computation 

}




void PALSol::
ResidComp()
{
	if(resdone) return; 			// Quick return.
	if(!eigdone || !iseigv) {
		erprint("Eigenpair not computed yet!");
	}

	res 		= new double[ nev ];

	/* Internal variables.				*/
	COMPLEX *temp 	= new COMPLEX[ n ];
	COMPLEX ev;
 	double 		relerr, t;

	/* Compute matrix 1-norm if necessary.		*/
	normm = ( normm==0 )? M.norm1() : normm;
	normd = ( normd==0 )? D.norm1() : normd;
	normk = ( normk==0 )? K.norm1() : normk;

	/* Calculate relative residual norms.		*/

	for ( int i=0; i<nev; i++ ) {
		cblas_zdscal( n, 0, temp, 1);		// Initialize to zero.
		ev = EigVal[i];				        // Eigenvalue i.
		
		M.MultMv(&EigVec[i*nl], temp, ev*ev);	// Compute residual Q(e_i)x_i.
		D.MultMv(&EigVec[i*nl], temp, ev);
		K.MultMv(&EigVec[i*nl], temp, 1);
		
		relerr = cblas_dznrm2(n, temp, 1);	// Residual 2-norm r = |Q(e_i)x_i|_2
  		
		res[i] = relerr/(ABS(ev)*ABS(ev)*normm+ABS(ev)*normd+normk);
	       						// Relative residual norm	
	}

	resdone = true; 	// Relative residual computed.

}// compute the relative residual norms




COMPLEX PALSol::
Eigenvalue( int i )
{
	if(!eigdone) EigComp(); 		// If the eigenvalues are not computed.

	if( i<0 || i>nconv ) {
		erprint("EIGENVALUE INDEX OUT OF RANGE");
		return 0;
	}
	else{ 
		return EigVal[i-1]; 
	}
}


void* PALSol::
Eigenvector(int i)
{
	if( !eigdone || !iseigv ){
		erprint("Eigenvector not computed yet!");
	} 

	if( i<0 || i>nconv ) {
		erprint("EIGENVECTOR INDEX OUT OF RANGE");
		return NULL;
	}
	else{ 
		return &EigVec[ nl * (i-1) ]; 
	}
}



double PALSol::
Resid( int i )
{
	if(!resdone) ResidComp();	// If the residual norm are not computed.

	if( i<0 || i>nconv ) {
		erprint("EIGENVALUE INDEX OUT OF RANGE");
		return 0;
	}
	else{ 
		return res[ i-1 ];
	}
}



void PALSol::
SetStat() // output static
{
	debug_.logfil = 6;
    	debug_.ndigit = -3;
    	debug_.mcaupd = 1;
    	debug_.mcaup2 = 0;
    	debug_.mcaitr = 0;
    	debug_.mceigt = 0;
    	debug_.mcapps = 0;
    	debug_.mcgets = 0;
    	debug_.mceupd = 0;
}



void PALSol::
PrintEig( int prec ) // Print eigenvalues and backward error.
{
	if(!resdone) ResidComp();	// Check if everything is computed.

	COMPLEX	t;

  	printf("\n\n%5s:\t (%*s, %*s) \t %6s \n","#Eig", 
			prec+7,"Real",prec+7,"Imag","BackErr");

	for(int i=1; i<=nconv; i++){
		t = Eigenvalue(i);
  		printf("%5i:\t (%+*.*E, %+*.*E)\t %2.1E \n",i, prec+1, prec,t.real(),
				prec+1, prec, t.imag(),Resid(i));
	}
}



void PALSol::
PrintEig() // Print eigenvalues with 5 decimal, and backward error.
{
	PrintEig(5); 
}

