#ifndef _PALSOL_H
#define _PALSOL_H

#include"slu_zdefs.h" 
#include"palmat.h"
#include"palar.h"
#include"paldef.h"
#include<string>
#include<iostream>

using namespace std;

/*
   -----------------------------------------------
   The PAL matrix class, matrix size being n+p*rk, 
   -----------------------------------------------
*/
class PALSol{

 private:

  /* a) Problem variables */

  int   p; 		    // Pade approximant order. 
  int	rk;		    // Rank of the damping matrix.
  COMPLEX sigma;    // shifts 
  PALMatrix M, D, K;// Matrix M, D, K and F. CSC format.


  int    nev;		// Number of eigenvalues to be computed.



  /* b) Internal variables */

    /* b.1) SUPERLU variables */

  double	threshold;	// Row pivoting threshold, default = 0.1.
  int		order;		// Ordering parameters, default = 1, natural order.
  int		*permc, *permr;	// Array of row and column permutation
  SuperMatrix	L, U;		// L, U factors
  SuperLUStat_t	stat;		// Variables.
  bool		isfactor;	// Indicates whether LU factorization is done.

  	/* b.2) PAL variables */

  LRMat 	E, F;	    // store the factorization D=EF

  int		n; 		    // Dimension of the QEP
  int		nl; 		// Dimension of the LEP
  double	delta;		// scaling factor.
  double 	*b;         // E = -diag(b), size p-by-p.
  double 	*a;	        // length p vector.
  double 	normm, normd, normk;   // matrix 1-norms.
  int  		nzmax;		// Max number of nonzero elements of Q(sigma)=sigma^2*M+sigma*D+K.
  				        // (default by nnzm+nnzd+nnzk)
  double 	rtol;		// Backward error tolerance of PAL.
  				        // (default by eps.)  
  bool 		isscaling; 	// true for applying scaling; false, otherwise.
  bool 		isscaled; 	// true for applying scaling; false, otherwise.
  bool  	iseigv;		// true if eigvectors required; false, otherwise.
  bool  	isresid;	// true if relative residual norms required; false, otherwise.
  bool  	resdone;	// true if relative residual norms required; false, otherwise.
  int		nconv;		// number of converged eigenvalues
  				        // set to -1 at the beginning
  int		*idxe;		// index of converged eigenvalues and eigenvectors

  	/* b.3) Arpack variables */

  double	artol;		// (default=eps) Stopping criteria of Arnoldi iteration.
  int	 	ncv;		// (default=2*nev+1) Number of Arnoldi vectors at each iteration.
  int	 	maxit;		// (default=300) Maximum number of iteration of IRAM allowed.
  bool		eigdone;	// Indicates whether LEP is solved.
  int		isv0;
 	/* b.4) Data variables */ 

  double 	*res;		// Eigenpair relative residual norms
  COMPLEX	*EigVal;	// Eigenvalues
  COMPLEX 	*EigVec;	// Eigenvectors



 public:

  int 	np() {return nl;}
  // function that returns the dimension of the problem

  bool 	parachk();
  // parameter checking

  double scaling();
  // Compute matrix norms and scaling

  void 	Qsigma(COMPLEX* val, int* colptr, int* rowind,  int& nnz, int& info);
  // Compute Q(sigma) = sigma^2M + sigma*D + K
  
  void 	lufactor();
  // compute lu factorization of Q(sigma)

  void 	MultMv(COMPLEX* v, COMPLEX* w);
  // Matrix vector multiplication w <- M*v.

  void 	EigComp();
  // compute eigenvalues by ARPACK
  
  //  void eigvec();
  //  compute eigenvectors

  void 	ResidComp();
  // compute residuals

  void*	Eigenvalue();
  // return eigenvalues

  COMPLEX Eigenvalue( int i);
  // return i-th eigenvalue

  //  void*	Eigenvector();
  // return eigenvectors, not support 
  
  void*	Eigenvector( int i);
  // return i-th eigenvector

  double Resid( int i);
  // return backward error of i-th eigenpair. 
 
  void PrintEig( int prec );
  void PrintEig( );
  // Print eigenvalues and backward error.

  //  void geteig();
  // retrieve eigenvalue and eigenvectors


  PALSol(int p, COMPLEX sigma, int nev, 
		  PALMatrix M, PALMatrix D, PALMatrix K);
 // Constructor

  /* Parameter handler. */
  void SetTol(double x) { rtol = (x>0)? x: 1.0E-16;}; 	// backward error tolerance: CAN BE IMPROVED
  void SetARTol(double x) { artol = (x>0)? x: 1.0E-16;};	// ARPACK stopping criteria
  void SetARMaxit(int x) { maxit = (x>300||x<0)? 300: x;};	// ARPACK stopping criteria
  void SetLUThresh(double x) { threshold = (x>1||x<0)? 1 : x;};
  void NoScaling() { isscaling = false;} 
  void NoEigVec() { iseigv = false;} 

  void SetStat(); // output static
  // ~PALSol() 

}; // ComplexMatrixE.

#endif
